# Adapted from
# https://github.com/numpy/numpy/blob/81d53e44a8d503dba3101c1f523008d0a0dde787/numpy/fft/tests/test_fftpack.py
# in order to ensure we have the same behaviour as the Numpy API
from __future__ import division, absolute_import, print_function

import numpy as np
from numpy.random import random
from numpy.testing import TestCase, run_module_suite, assert_array_almost_equal
from numpy.testing import assert_array_equal
import threading
import sys
if sys.version_info[0] >= 3:
    import queue
else:
    import Queue as queue

import mklfft.fftpack as mklfft

def fft1(x):
    L = len(x)
    phase = -2j*np.pi*(np.arange(L)/float(L))
    phase = np.arange(L).reshape(-1, 1) * phase
    return np.sum(x*np.exp(phase), axis=1)


class TestFFTShift(TestCase):

    def test_fft_n(self):
        self.assertRaises(ValueError, mklfft.fft, [1, 2, 3], 0)


class TestFFT1D(TestCase):

    def test_fft(self):
        x = random(30) + 1j*random(30)
        assert_array_almost_equal(fft1(x), mklfft.fft(x))
        assert_array_almost_equal(fft1(x) / np.sqrt(30),
                                  mklfft.fft(x, norm="ortho"))

    def test_ifft(self):
        x = random(30) + 1j*random(30)
        assert_array_almost_equal(x, mklfft.ifft(mklfft.fft(x)))
        assert_array_almost_equal(
            x, mklfft.ifft(mklfft.fft(x, norm="ortho"), norm="ortho"))

    def test_fft2(self):
        x = random((30, 20)) + 1j*random((30, 20))
        assert_array_almost_equal(mklfft.fft(mklfft.fft(x, axis=1), axis=0),
                                  mklfft.fft2(x))
        assert_array_almost_equal(mklfft.fft2(x) / np.sqrt(30 * 20),
                                  mklfft.fft2(x, norm="ortho"))

    def test_ifft2(self):
        x = random((30, 20)) + 1j*random((30, 20))
        assert_array_almost_equal(mklfft.ifft(mklfft.ifft(x, axis=1), axis=0),
                                  mklfft.ifft2(x))
        assert_array_almost_equal(mklfft.ifft2(x) * np.sqrt(30 * 20),
                                  mklfft.ifft2(x, norm="ortho"))

    def test_fftn(self):
        x = random((30, 20, 10)) + 1j*random((30, 20, 10))
        assert_array_almost_equal(
            mklfft.fft(mklfft.fft(mklfft.fft(x, axis=2), axis=1), axis=0),
            mklfft.fftn(x))
        assert_array_almost_equal(mklfft.fftn(x) / np.sqrt(30 * 20 * 10),
                                  mklfft.fftn(x, norm="ortho"))

    def test_ifftn(self):
        x = random((30, 20, 10)) + 1j*random((30, 20, 10))
        assert_array_almost_equal(
            mklfft.ifft(mklfft.ifft(mklfft.ifft(x, axis=2), axis=1), axis=0),
            mklfft.ifftn(x))
        assert_array_almost_equal(mklfft.ifftn(x) * np.sqrt(30 * 20 * 10),
                                  mklfft.ifftn(x, norm="ortho"))

    def test_rfft(self):
        x = random(30)
        assert_array_almost_equal(mklfft.fft(x)[:16], mklfft.rfft(x))
        assert_array_almost_equal(mklfft.rfft(x) / np.sqrt(30),
                                  mklfft.rfft(x, norm="ortho"))

    def test_irfft(self):
        x = random(30)
        assert_array_almost_equal(x, mklfft.irfft(mklfft.rfft(x)))
        assert_array_almost_equal(
            x, mklfft.irfft(mklfft.rfft(x, norm="ortho"), norm="ortho"))

    def test_rfft2(self):
        x = random((30, 20))
        assert_array_almost_equal(mklfft.fft2(x)[:, :11], mklfft.rfft2(x))
        assert_array_almost_equal(mklfft.rfft2(x) / np.sqrt(30 * 20),
                                  mklfft.rfft2(x, norm="ortho"))

    def test_irfft2(self):
        x = random((30, 20))
        assert_array_almost_equal(x, mklfft.irfft2(mklfft.rfft2(x)))
        assert_array_almost_equal(
            x, mklfft.irfft2(mklfft.rfft2(x, norm="ortho"), norm="ortho"))

    def test_rfftn(self):
        x = random((30, 20, 10))
        assert_array_almost_equal(mklfft.fftn(x)[:, :, :6], mklfft.rfftn(x))
        assert_array_almost_equal(mklfft.rfftn(x) / np.sqrt(30 * 20 * 10),
                                  mklfft.rfftn(x, norm="ortho"))

    def test_irfftn(self):
        x = random((30, 20, 10))
        assert_array_almost_equal(x, mklfft.irfftn(mklfft.rfftn(x)))
        assert_array_almost_equal(
            x, mklfft.irfftn(mklfft.rfftn(x, norm="ortho"), norm="ortho"))

    def test_hfft(self):
        x = random(14) + 1j*random(14)
        x_herm = np.concatenate((random(1), x, random(1)))
        x = np.concatenate((x_herm, x[::-1].conj()))
        assert_array_almost_equal(mklfft.fft(x), mklfft.hfft(x_herm))
        assert_array_almost_equal(mklfft.hfft(x_herm) / np.sqrt(30),
                                  mklfft.hfft(x_herm, norm="ortho"))

    def test_ihttf(self):
        x = random(14) + 1j*random(14)
        x_herm = np.concatenate((random(1), x, random(1)))
        x = np.concatenate((x_herm, x[::-1].conj()))
        assert_array_almost_equal(x_herm, mklfft.ihfft(mklfft.hfft(x_herm)))
        assert_array_almost_equal(
            x_herm, mklfft.ihfft(mklfft.hfft(x_herm, norm="ortho"),
                                 norm="ortho"))


class TestFFTThreadSafe(TestCase):
    threads = 16
    input_shape = (800, 200)

    def _test_mtsame(self, func, *args):
        def worker(args, q):
            q.put(func(*args))

        q = queue.Queue()
        expected = func(*args)

        # Spin off a bunch of threads to call the same function simultaneously
        t = [threading.Thread(target=worker, args=(args, q))
             for i in range(self.threads)]
        [x.start() for x in t]

        [x.join() for x in t]
        # Make sure all threads returned the correct value
        for i in range(self.threads):
            assert_array_equal(q.get(timeout=5), expected,
                'Function returned wrong value in multithreaded context')

    def test_fft(self):
        a = np.ones(self.input_shape) * 1+0j
        self._test_mtsame(mklfft.fft, a)

    def test_ifft(self):
        a = np.ones(self.input_shape) * 1+0j
        self._test_mtsame(mklfft.ifft, a)

    def test_rfft(self):
        a = np.ones(self.input_shape)
        self._test_mtsame(mklfft.rfft, a)

    def test_irfft(self):
        a = np.ones(self.input_shape) * 1+0j
        self._test_mtsame(mklfft.irfft, a)


if __name__ == "__main__":
    run_module_suite()
